from heuristicsearch.graph import Graph 
class AOStar(Graph):
    def __init__(self, graph, heuristicNodeList, startNode):  #instantiate graph object with graph topology, heuristic values, start node
        super().__init__(graph)
        self.__H = heuristicNodeList
        self.__start = startNode
        self.__parent = {}
        self.__status = {}
        self.__solutionGraph = {}

    def applyAOStar(self):  # starts a recursive AO* algorithm
        self.__aoStar(self.__start, False)
        self.__printSolution()

    def __getNeighbors(self, v):  # gets the Neighbors of a given node
        return self.adjac_lis.get(v, '')

    def __getStatus(self, v):  # return the status of a given node
        return self.__status.get(v, 0)

    def __setStatus(self, v, val):  # set the status of a given node
        self.__status[v] = val

    def __getHeuristicNodeValue(self, n):
        return self.__H.get(n, 0)  # always return the heuristic value of a given node

    def __setHeuristicNodeValue(self, n, value):
        self.__H[n] = value  # set the revised heuristic value of a given node

    def __printSolution(self):
        print("FOR THE SOLUTION, TRAVERSE THE GRAPH FROM THE START NODE:", self.__start)
        print("------------------------------------------------------------")
        print(self.__solutionGraph)
        print("------------------------------------------------------------")

    def __computeMinimumCostChildNodes(self, v):  # Computes the Minimum Cost of child nodes of a given node v
        minimumCost = 0
        costToChildNodeListDict = {}
        costToChildNodeListDict[minimumCost] = []
        flag = True
        for nodeInfoTupleList in self.__getNeighbors(v):  # iterate over all the set of child node/s
            cost = 0
            nodeList = []
            for c, weight in nodeInfoTupleList:
                cost = cost + self.__getHeuristicNodeValue(c) + weight
                nodeList.append(c)
            if flag == True:  # initialize Minimum Cost with the cost of first set of child node/s
                minimumCost = cost
                costToChildNodeListDict[minimumCost] = nodeList  # set the Minimum Cost child node/s
                flag = False
            else:  # checking the Minimum Cost nodes with the current Minimum Cost
                if minimumCost > cost:
                    minimumCost = cost
                    costToChildNodeListDict[minimumCost] = nodeList  # set the Minimum Cost child node/s
        return minimumCost, costToChildNodeListDict[minimumCost]  # return Minimum Cost and Minimum Cost child node/s

    def __aoStar(self, v, backTracking):  # AO* algorithm for a start node and backTracking status flag
        # print("HEURISTIC VALUES :", self.__H)
        # print("SOLUTION GRAPH :", self.__solutionGraph)
        print("PROCESSING NODE :", v)
        print("-----------------------------------------------------------------------------------------")
        if self.__getStatus(v) >= 0:  # if status node v >= 0, compute Minimum Cost nodes of v
            minimumCost, childNodeList = self.__computeMinimumCostChildNodes(v)
            print(minimumCost, childNodeList)
            print()
            self.__setHeuristicNodeValue(v, minimumCost)
            self.__setStatus(v, len(childNodeList))
            solved = True  # check the Minimum Cost nodes of v are solved
            for childNode in childNodeList:
                self.__parent[childNode] = v
                if self.__getStatus(childNode) != -1:
                    solved = solved & False
            if solved == True:  # if the Minimum Cost nodes of v are solved, set the current node status as solved(-1)
                self.__setStatus(v, -1)
                self.__solutionGraph[
                    v] = childNodeList  # update the solution graph with the solved nodes which may be a part of solution
            if v != self.__start:  # check the current node is the start node for backtracking the current node value
                self.__aoStar(self.__parent[v],
                            True)  # backtracking the current node value with backtracking status set to true
            if backTracking == False:  # check the current call is not for backtracking
                for childNode in childNodeList:  # for each Minimum Cost child node
                    self.__setStatus(childNode, 0)  # set the status of child node to 0(needs exploration)
                    self.__aoStar(childNode,
                                False)  # Minimum Cost child node is further explored with backtracking status as false



from heuristicsearch.graph import Graph

class AStar(Graph):
    def __init__(self, adjac_lis, heuristicNodeList):
        super().__init__(adjac_lis)
        self.heuristic = heuristicNodeList
    def __get_neighbors(self, v):
        return self.adjac_lis[v]
 
    def __get_heuristic(self, n):
        return self.heuristic[n]
    def apply_a_star(self, start, stop):
        self.__a_star_algorithm(start=start, stop=stop)
    
    def __a_star_algorithm(self, start, stop):
        # In open_lst is a list of nodes which have been visited, but who's 
        # neighbours haven't all been inspected, It starts off with the start node
        # And closed_lst is a list of nodes which have been visited
        # and who's neighbors have been inspected
        open_lst = set([start])
        closed_lst = set([])
 
        # poo has present distances from start to all other nodes
        # the default value is +infinity
        poo = {start:0}
 
        # par contains an adjac mapping of all nodes
        par = {start:start}
 
        while len(open_lst) > 0:
            n = None
 
            # it will find a node with the lowest value of f(n) = g(n) + h(n)
            for v in open_lst:
                if n == None or poo[v] + self.__get_heuristic(v) < poo[n] + self.__get_heuristic(n):
                    n = v
 
            if n == None:
                print('Path does not exist!')
                return None
 
            # if the current node is the stop
            # then we start again from start
            if n == stop:
                reconst_path = []
 
                while par[n] != n:
                    reconst_path.append(n)
                    n = par[n]
 
                reconst_path.append(start)
 
                reconst_path.reverse()
                print("Path")
                for i in range(len(reconst_path) - 1):
                    print(f"{reconst_path[i]} -> ",end="")
                print(reconst_path[-1])
                print("Cost")
                for i in range(len(reconst_path) - 1):
                    print(f"{poo[reconst_path[i]]} -> ",end="")
                print(poo[reconst_path[-1]])
                # print('Path found: {} and the total cost is {}'.format(reconst_path,poo[stop]))
                return reconst_path
 
            # for all the neighbors of the current node do
            for (m, weight) in self.__get_neighbors(n):
              # if the current node is not presentin both open_lst and closed_lst
                # add it to open_lst and note n as it's par
                if m not in open_lst and m not in closed_lst:
                    open_lst.add(m)
                    par[m] = n
                    poo[m] = poo[n] + weight
 
                # otherwise, check if it's quicker to first visit n, then m
                # and if it is, update par data and poo data
                # and if the node was in the closed_lst, move it to open_lst
                else:
                    if poo[m] > poo[n] + weight:
                        poo[m] = poo[n] + weight
                        par[m] = n
 
                        if m in closed_lst:
                            closed_lst.remove(m)
                            open_lst.add(m)
 
            # remove n from the open_lst, and add it to closed_lst
            # because all of his neighbors were inspected
            open_lst.remove(n)
            closed_lst.add(n)
 
        print('Path does not exist!')
        return None